use ddc_hi::Display;
use paris::{info, error};
use tokio::sync::{broadcast, OnceCell};

use crate::{display::Brightness, rpc};

use super::DisplayBrightness;
use ddc_hi::Ddc;

pub struct Manager {}

impl Manager {
    pub async fn global() -> &'static Self {
        static DISPLAY_MANAGER: OnceCell<Manager> = OnceCell::const_new();

        DISPLAY_MANAGER.get_or_init(|| Self::create()).await
    }

    pub async fn create() -> Self {
        let rpc = rpc::Manager::global().await;

        let rx = rpc.client().subscribe_change_display_brightness_rx();
        tokio::spawn(Self::subscribe_display_brightness(rx));

        Self {}
    }

    pub async fn subscribe_display_brightness(mut rx: broadcast::Receiver<DisplayBrightness>) {
        while let Ok(display_brightness) = rx.recv().await {
            if let Err(err) = Self::set_display_brightness(display_brightness) {
                error!("set_display_brightness failed. {:?}", err);
            }
        }
    }

    pub fn set_display_brightness(display_brightness: DisplayBrightness) -> anyhow::Result<()> {
        match Display::enumerate().get_mut(display_brightness.display_index) {
            Some(display) => match display.handle.get_vcp_feature(0x10) {
                Ok(curr_brightness) => {
                    let curr = curr_brightness.value();
                    info!("curr_brightness: {:?}", curr);
                    let mut target = match display_brightness.brightness {
                        Brightness::Relative(v) => v + curr,
                        Brightness::Absolute(v) => v,
                    };
                    if target.gt(&curr_brightness.maximum()) {
                        target = curr_brightness.maximum();
                    } else if target.lt(&0) {
                        target = 0;
                    }
                    display
                        .handle
                        .set_vcp_feature(0x10, target)
                        .map_err(|err| anyhow::anyhow!("can not set brightness. {:?}", err))?;
                }
                Err(err) => {
                    info!(
                        "can not get display#{} brightness. {:?}",
                        display_brightness.display_index, err
                    );
                    if let Brightness::Absolute(v) = display_brightness.brightness {
                        display
                            .handle
                            .set_vcp_feature(0x10, v)
                            .map_err(|err| anyhow::anyhow!("can not set brightness. {:?}", err))?;
                    };
                }
            },
            None => {
                anyhow::bail!(
                    "display#{} was not found.",
                    display_brightness.display_index
                );
            }
        }
        Ok(())
    }
}
